from debian import substvars
import os
import pathlib
import unittest
import shutil
import sys
import tempfile


def find_pkg_root():
    if not os.environ.get("PKG_KDE_TOOLS_TEST_INSTALLED", False):
        for p in pathlib.Path(__file__).parents:
            if (p/"pkgkde-getqmldepends").exists():
                return p
    return pathlib.Path("/usr/share/pkg-kde-tools")

sys.path.insert(0, str(find_pkg_root()))

from pythonlib import pkgkde

class QmlDependsTestCase(unittest.TestCase):
    def check_substvar(self, fieldname, expected, pkg_name="testpackage"):
        with substvars.Substvars.load_from_path(f"debian/{pkg_name}.substvars") as svars:
            self.assertEqual(svars.get(fieldname, ""), expected)

    def writeControlFile(self):
        with (pathlib.Path(self.tempdir.name)/"debian/control").open("wb") as f:
            self.dc.source.dump(f)
            for pkg in self.dc.packages.values():
                f.write(b"\n")
                pkg.dump(f)

    def setUp(self):
        self.old_cwd = os.getcwd()
        self.tempdir = tempfile.TemporaryDirectory()
        shutil.copytree(pathlib.Path(__file__).with_name("qmltestpackage"), self.tempdir.name, dirs_exist_ok=True)
        os.chdir(self.tempdir.name)
        pathlib.Path("debian/testpackage.substvars").unlink(missing_ok=True)
        self.dc = pkgkde.DebianControl(pathlib.Path("debian/control"))

    def tearDown(self):
        self.tempdir.cleanup()
        os.chdir(self.old_cwd)





#!/usr/bin/env python3
# SPDX-FileCopyrightText: 2023 Sandro Knauß <hefee@debian.org>
# SPDX-License-Identifier: LGPL-2.1-or-later

import contextlib
import json
import logging
import pathlib
import re
import subprocess

from debian import deb822, substvars

logging.basicConfig(format='%(levelname).1s: pkgkde '
                           '%(module)s:%(lineno)d: %(message)s')
log = logging.getLogger('pkgkde')

PATHS = [pathlib.Path(__file__).parent.with_name("datalib"),
         pathlib.Path('/usr/share/pkg-kde-tools/lib')
        ]

deb822_logger = logging.getLogger('debian.deb822')
REL_RAW_RE = re.compile('cannot parse package relationship ".*", returning it raw')

class DebianControl:
    """ Simple class for debian/control file.

    source: with point to the source entry in contrile file
    packages: dict with all binary packages in control file

    """
    def __init__(self, path: pathlib.Path) -> None:
        self.path = path

        self.read_path(path)

    def read_path(self, path:pathlib.Path) -> None:
        self.source = None
        self.packages = {}
        with path.open() as cl:
            for block in deb822.Deb822.iter_paragraphs(cl):
                if block.get("Source"):
                    self.source = block
                else:
                    self.packages[block.get("Package")] = block

    def isSinglePackage(self) -> bool:
        """Is there only one binary package?"""
        return len(self.packages) == 1


def active_compat_level() -> int:
    output = subprocess.check_output(["dh_assistant", "active-compat-level"])
    j = json.loads(output)
    return j["active-compat-level"]


def set_substvar(path:pathlib.Path, substvar_dict: dict):
    """ handles a debian/substvars file.

        path: path to substvars file.
        substsvars: is a dict to set key,value in the substvars file.
    """

    with substvars.Substvars.load_from_path(path, missing_ok=True) as svars:
        for var_name, values in substvar_dict.items():
            for pkg in values:
                svars.add_dependency(var_name, pkg)


def filter_rel_raw_warning(record):
    return record.levelno != logging.WARNING or not REL_RAW_RE.match(record.getMessage())


@contextlib.contextmanager
def deb822_filter_rel_raw_warning():
    deb822_logger.addFilter(filter_rel_raw_warning)
    try:
        yield
    finally:
        deb822_logger.removeFilter(filter_rel_raw_warning)



#!/usr/bin/python3

import pathlib
import unittest
import sys

from qmldependstestcase import QmlDependsTestCase

from pythonlib import qmldeps

class TestQmlDepends(QmlDependsTestCase):
    def test_var_missing_in_control(self):
        pkg = self.dc.packages["testpackage"]
        for field in ("qml6:depends", "qml6:Recommends", "qml6:Suggests"):
            with self.subTest(field=field):
                self.assertEqual(qmldeps.is_var_missing_in_control(pkg, field), True)

        pkg["Suggests"] = "${qml6:Depends}, ${qml6:Recommends}, ${qml6:Suggests}"
        for field in ("qml6:Depends", "qml6:Recommends", "qml6:Suggests"):
            with self.subTest(suggests=True, field=field):
                self.assertEqual(qmldeps.is_var_missing_in_control(pkg, field), False)

        pkg["Depends"] = pkg["Suggests"]
        del pkg["Suggests"]
        for field in ("qml6:Depends", "qml6:Recommends", "qml6:Suggests"):
            with self.subTest(depends=True, field=field):
                self.assertEqual(qmldeps.is_var_missing_in_control(pkg, field), False)

    def test_var_complete_demoted(self):
        pkg = self.dc.packages["testpackage"]
        for f in ("Depends", "Recommends", "Suggests"):
            for var_name in ("qml6:depends", "qml6:Recommends", "qml6:Suggests"):
                with self.subTest(field=f, var_name=var_name):
                    self.assertEqual(qmldeps.is_var_complete_demoted(pkg, var_name, f), False)

        pkg["Suggests"] = "${qml6:Depends}, ${qml6:Recommends}, ${qml6:Suggests}"
        for f in ("Depends", "Recommends", "Suggests"):
            for var_name in ("qml6:Depends", "qml6:Recommends", "qml6:Suggests"):
                with self.subTest(suggests=True, field=f, var_name=var_name):
                    self.assertEqual(qmldeps.is_var_complete_demoted(pkg, var_name, f), f!="Suggests")

        pkg["Depends"] = pkg["Suggests"]
        del pkg["Suggests"]
        for f in ("Depends", "Recommends", "Suggests"):
            for var_name in ("qml6:Depends", "qml6:Recommends", "qml6:Suggests"):
                with self.subTest(depends=True, field=f, var_name=var_name):
                    self.assertEqual(qmldeps.is_var_complete_demoted(pkg, var_name, f), False)

    def test_partial_demoted(self):
        pkg = self.dc.packages["testpackage"]
        for f in ("Depends", "Recommends", "Suggests"):
            dependencies = set(("qml6-module-blub",))
            with self.subTest(field=f):
                self.assertEqual(qmldeps.partial_demotion(pkg, dependencies ,f), False)
                self.assertEqual(dependencies, set(("qml6-module-blub",)))

        pkg["Suggests"] = "qml6-module-blub, ignored-dep"
        for f in ("Depends", "Recommends", "Suggests"):
            dependencies = set(("qml6-module-blub", "unaffected-dep"))
            with self.subTest(suggests=True,field=f):
                if f == "Suggests":
                    self.assertEqual(qmldeps.partial_demotion(pkg, dependencies ,f), False)
                    self.assertEqual(dependencies, set(("qml6-module-blub","unaffected-dep")))
                else:
                    with self.assertLogs('dh_qmldeps', level='WARNING') as cm:
                        self.assertEqual(qmldeps.partial_demotion(pkg, dependencies ,f), True)
                    self.assertEqual(cm.output,[f"WARNING:dh_qmldeps:Detected dependencies ({{'qml6-module-blub'}}) are demoted from {f} to Suggests. To be future proof with debhelper level 15, add -d Suggests with the corresponding qml files."])
                    self.assertEqual(dependencies, set(("unaffected-dep",)))  # Demotion happend

    def test_partial_promition(self):
        pkg = self.dc.packages["testpackage"]
        pkg["Depends"] = "qml6-module-blub, ignored-dep"
        for f in ("Depends", "Recommends", "Suggests"):
            dependencies = set(("qml6-module-blub",))
            with self.subTest(depends=True, field=f):
                self.assertEqual(qmldeps.partial_demotion(pkg, dependencies ,f), False)
                self.assertEqual(dependencies, set(("qml6-module-blub",)))


class TestQmlDependsDH15(QmlDependsTestCase):
    dh_compat = "15"
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.create_no_context = False
        self.verbose = "-v" in sys.argv

    def setUp(self):
        super().setUp()
        self.create_no_context = False
        if self.dc.source["X-DH-Compat"] != self.dh_compat:
            self.dc.source["X-DH-Compat"] = self.dh_compat
            self.writeControlFile()

    def _call_main(self, args):
        if self.verbose:
            args = ["-v",] + args
        return qmldeps.main(args)

    def _test_depends(self):
        self._call_main(["-d", "Depends", "-p", "testpackage", "--", "-qmlFiles", "quick.qml"])
        self.check_substvar("qml6:Depends", "qml6-module-qtquick")

    test_depends = _test_depends
    _test_partial_depends = _test_depends

    def test_recommends(self):
        self._call_main(["-d", "Recommends", "-p", "testpackage", "--", "-qmlFiles", "quick.qml"])
        self.check_substvar("qml6:Recommends", "qml6-module-qtquick")

    _test_recommends = test_recommends

    def test_suggests(self):
        self._call_main(["-d", "Suggests", "-p", "testpackage", "--", "-qmlFiles", "quick.qml"])
        self.check_substvar("qml6:Suggests", "qml6-module-qtquick")

    def test_multiplecalls(self):
        self._call_main(["-d", "Depends", "-p", "testpackage", "--", "-qmlFiles", "quick.qml"])
        self._call_main(["-d", "Suggests", "-p", "testpackage", "--", "-qmlFiles", "test.qml"])
        self._call_main(["-d", "Recommends", "-p", "testpackage", "--", "-qmlFiles", "both.qml"])
        self.check_substvar("qml6:Depends", "qml6-module-qtquick")
        self.check_substvar("qml6:Suggests", "qml6-module-qttest")
        self.check_substvar("qml6:Recommends", "qml6-module-qtquick, qml6-module-qttest")

    def test_full_demotion(self):
        """ On Debhelper > 13 full demotion is ignored"""
        pkg = self.dc.packages["testpackage"]
        pkg["Suggests"] = "${qml6:Depends}"
        self.writeControlFile()
        self._test_depends()

    def test_only_depends_var_usage_preor_0_22(self):
        # testcase prior 0.22 version only qml6:Depends exists and users used
        # it in other fields. Make sure that the build does not break, but only
        # list an error for dh 13.
        pkg = self.dc.packages["testpackage"]
        pkg["Suggests"] = "${qml6:Depends}"
        self.writeControlFile()
        self._test_recommends()

    def test_partial_demotion(self):
        """ On Debhelper > 14 partial demotion is ignored"""
        pkg = self.dc.packages["testpackage"]
        pkg["Suggests"] = "qml6-module-qtquick"
        self.writeControlFile()
        self._test_partial_depends()

    def test_override(self):
        (pathlib.Path(self.tempdir.name)/"debian/qmldeps.overrides").write_text("QtQuick foo\n")
        self._call_main(["-d", "Depends", "-p", "testpackage", "--", "-qmlFiles", "quick.qml"])
        self.check_substvar("qml6:Depends", "foo")

    def test_optional_no_change(self):
        (pathlib.Path(self.tempdir.name)/"debian/qmldeps.optional").write_text("QtQuick foo\n")
        self._call_main(["-d", "Depends", "-p", "testpackage", "--", "-qmlFiles", "quick.qml"])
        self.check_substvar("qml6:Depends", "qml6-module-qtquick")

    def test_optional_missing(self):
        args = ["-d", "Depends", "-p", "testpackage", "--", "-qmlFiles", "missing.qml"]
        with self.assertLogs('dh_qmldeps', level='ERROR') as cm:
            with self.assertRaises(SystemExit):
                self._call_main(args)
        self.assertEqual(cm.output,["ERROR:dh_qmldeps:Missing QML module IAmMissing",])

        (pathlib.Path(self.tempdir.name)/"debian/qmldeps.optional").write_text("IAmMissing foo\n")
        self._call_main(args)
        self.check_substvar("qml6:Depends", "foo")


class TestQmlDependsDH14(TestQmlDependsDH15):
    dh_compat = "14"

    def partial_demotion_msg(self, deps, _from, _to):
        return f"WARNING:dh_qmldeps:Detected dependencies ({deps}) are demoted from {_from} to {_to}. To be future proof with debhelper level 15, add -d {_to} with the corresponding qml files."

    def _test_partial_depends(self):
        """qml6-module-qtquick is demoted to suggests"""
        args = ["-d", "Depends", "-p", "testpackage", "--", "-qmlFiles", "quick.qml"]
        if not self.create_no_context:
            with self.assertLogs('dh_qmldeps', level='WARNING') as cm:
                self._call_main(args)
            self.assertEqual(cm.output,[self.partial_demotion_msg(set(("qml6-module-qtquick",)),"Depends", "Suggests")])
        else:
            self._call_main(args)
        self.check_substvar("qml6:Depends", "")


class TestQmlDependsDH13(TestQmlDependsDH14):
    dh_compat = "13"

    def missing_field(self, fieldname):
        return f"ERROR:dh_qmldeps:cannot insert qml dependency (missing qml6:{fieldname} in Depends/Recommends/Suggests)"

    def check_error_call(self, func, fieldname):
        with self.assertLogs('dh_qmldeps', level='ERROR') as cm:
            with self.assertRaises(SystemExit):
                self.create_no_context = True
                func()
        self.assertEqual(cm.output,[self.missing_field(fieldname),])
        self.create_no_context = False

    def check_call(self, func, fieldname):
        self.check_error_call(func, fieldname)

        pkg = self.dc.packages["testpackage"]
        pkg[fieldname] = f"${{qml6:{fieldname}}}"
        self.writeControlFile()

        func()

    def test_depends(self):
        self.check_call(super().test_depends, "Depends")

    def test_recommends(self):
        self.check_call(super().test_recommends, "Recommends")

    def test_suggests(self):
        self.check_call(super().test_suggests, "Suggests")

    def test_multiplecalls(self):
        self.check_error_call(super().test_multiplecalls, "Depends")

        pkg = self.dc.packages["testpackage"]
        pkg["Depends"] = "${qml6:Depends}"
        pkg["Recommends"] = "${qml6:Recommends}"
        pkg["Suggests"] = "${qml6:Suggests}"
        self.writeControlFile()
        super().test_multiplecalls()

    def test_full_demotion(self):
        with self.assertLogs('dh_qmldeps', level='WARNING') as cm:
            super().test_full_demotion()
        self.assertEqual(cm.output, ['WARNING:dh_qmldeps:The full demotion will not work anymore with Debhelper 14. Please use -d parameter to tell dh_qmldeps the correct dependency fieldname.'])

    def test_partial_demotion(self):
        self.check_call(super().test_partial_demotion, "Depends")

    def test_override(self):
        self.check_call(super().test_override, "Depends")

    def test_optional_no_change(self):
        self.check_call(super().test_optional_no_change, "Depends")

    def test_optional_missing(self):
        fieldname = "Depends"
        self.check_error_call(super().test_optional_missing, fieldname)

        pkg = self.dc.packages["testpackage"]
        pkg[fieldname] = f"${{qml6:{fieldname}}}"
        self.writeControlFile()

        (pathlib.Path(self.tempdir.name)/"debian/qmldeps.optional").unlink()

        super().test_optional_missing()

    def test_only_depends_var_usage_preor_0_22(self):
        with self.assertLogs('dh_qmldeps', level='WARNING') as cm:
            super().test_only_depends_var_usage_preor_0_22()
        self.assertEqual(cm.output,["WARNING:dh_qmldeps:qml6:Recommends is missing in Depends/Recommends/Suggests. pkg-kde-tools >= 0.22.0 supports mutliple fields. In order to not break builds, the detected dependencies are added to qml6:Depends, too."])
        self.check_substvar("qml6:Depends", "qml6-module-qtquick")
if __name__ == "__main__":
    unittest.main(verbosity=2)
